"""
You're writing code that ultimately needs to create a new class object. You've
thought about emitting class source code to a string and using a function
such as exec() to evaluate it, but you'd prefer a more elegant solution.
"""

import sys
import types
import operator


def example_1():
    # Example of making a class manually from parts
    # Methods
    def __init__(self, name, shares, price):
        self.name = name
        self.shares = shares
        self.price = price

    def cost(self):
        return self.shares * self.price

    cls_dict = {
        "__init__": __init__,
        "cost": cost,
    }

    # Make a class
    Stock = types.new_class("Stock", (), {}, lambda ns: ns.update(cls_dict))

    s = Stock("ACME", 50, 91.1)
    print(s)
    print(s.cost())


def example_2():
    # An alternative formulation of namedtuples
    def named_tuple(classname, fieldnames):
        # Populate a dictionary of field property accessors
        cls_dict = {
            name: property(operator.itemgetter(n)) for n, name in enumerate(fieldnames)
        }

        # Make a __new__ function and add to the class dict
        def __new__(cls, *args):
            if len(args) != len(fieldnames):
                raise TypeError("Expected {} arguments".format(len(fieldnames)))
            return tuple.__new__(cls, (args))

        cls_dict["__new__"] = __new__

        # Make the class
        cls = types.new_class(classname, (tuple,), {}, lambda ns: ns.update(cls_dict))
        cls.__module__ = sys._getframe(1).f_globals["__name__"]
        return cls

    Point = named_tuple("Point", ["x", "y"])
    print(Point)
    p = Point(4, 5)
    print(len(p))
    print(p.x, p[0])
    print(p.y, p[1])
    try:
        p.x = 2
    except AttributeError as e:
        print(e)
    print("%s %s" % p)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
Some bored script kiddie has entered the text 'pýtĥöñ' into a form
on your web page and you'd like to clean it up somehow.
"""
import unicodedata
import sys


def main():
    s = "pýtĥöñ\fis\tawesome\r\n"
    print("s:", s)

    remap = {ord("\t"): " ", ord("\f"): " ", ord("\r"): None}
    print("remap:", remap)

    a = s.translate(remap)
    print("a:", a)

    combined_chars = dict.fromkeys(
        c for c in range(sys.maxunicode) if unicodedata.combining(chr(c))
    )
    print("combined_chars: ", str(combined_chars)[:55], "...")

    b = unicodedata.normalize("NFD", a)
    print("b:", b)
    print("b.translate:", b.translate(combined_chars))
    print("encoded ascii:", b.encode("ascii", "ignore").decode("ascii"))

    digitmap = {
        c: ord("0") + unicodedata.digit(chr(c))
        for c in range(sys.maxunicode)
        if unicodedata.category(chr(c)) == "Nd"
    }
    print("digitmap:", str(digitmap)[:56], "...")
    print("len digitmap:", len(digitmap))

    x = "\u0661\u0662\u0663"
    print("x:", x)
    print("translated:", x.translate(digitmap))


if __name__ == "__main__":
    main()



"""
You need to convert or output integers represented by binary, octal
or hexidecimal integers.
"""


def example_1():
    x = 1234
    print("x:", x)

    print("\nbin:", bin(x))
    print("alt:", format(x, "b"))

    print("\noct:", oct(x))
    print("alt:", format(x, "o"))

    print("\nhex:", hex(x))
    print("alt:", format(x, "x"))
    print("=" * 20)


def example_2():
    x = -1234
    print("x:", x)
    print("bin (signed)", format(x, "b"))
    print("hex (signed)", format(x, "x"))
    print("\n32-bit unsigned bin:", format(2**32 + x, "b"))
    print("32-bit unsigned hex:", format(2**32 + x, "x"))
    print("=" * 20)


def example_3():
    val = "4d2"
    print("val:", val)
    print("val as int:", int(val, 16))

    val = "10011010010"
    print("val:", val)
    print("val as int:", int(val, 2))


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You need to read or write text data, possibly in different text encodings
such as ASCII, UTF-8 or UTF-16.
"""

from pathlib import Path


def main():
    # Some examples of reading text files with different options
    #
    # The file sample.txt is a UTF-8 encoded text file with Windows
    # line-endings (\r\n).

    # (a) Reading a basic text file (UTF-8 default encoding)
    sample_file = Path(__file__).parent / "sample.txt"

    print("Reading a simple text file (UTF-8)")
    with open(sample_file, "rt") as f:
        for line in f:
            print(repr(line))

    # (b) Reading a text file with universal newlines turned off
    print("Reading text file with universal newlines off")
    with open(sample_file, "rt", newline="") as f:
        for line in f:
            print(repr(line))

    # (c) Reading text file as ASCII with replacement error handling
    print("Reading text as ASCII with replacement error handling")
    with open(sample_file, "rt", encoding="ascii", errors="replace") as f:
        for line in f:
            print(repr(line))

    # (d) Reading text file as ASCII with ignore error handling
    print("Reading text as ASCII with ignore error handling")
    with open(sample_file, "rt", encoding="ascii", errors="ignore") as f:
        for line in f:
            print(repr(line))


if __name__ == "__main__":
    main()



import re
import textwrap
from typing import Mapping
import click
from colorama import Fore
from colorama.ansi import AnsiFore
from pyrecipes.chapter import Chapter
from pyrecipes.recipe import Recipe, SearchMatch
from .text import text_border


def render_recipe(recipe: Recipe, describe: bool = False):
    click.echo(recipe)
    if describe:
        click.echo(
            Fore.YELLOW
            + textwrap.fill(
                recipe.get_docstring(),
                70,
                initial_indent="   ",
                subsequent_indent="   ",
            )
            + Fore.RESET
            + "\n"
        )


def render_chapter(chapter: Chapter, describe: bool = False):
    click.echo(text_border(str(chapter), side_symbol=" "))
    for recipe in chapter:
        render_recipe(recipe, describe)
    click.echo("")


def get_n_matches(matches):
    count = 0
    for recipes in matches.values():
        for matches in recipes.values():
            count += sum(match.count for match in matches)
    return count


def render_match(
    pattern: str, match: SearchMatch, color: AnsiFore, ignore_case: bool = False
):
    flags = re.IGNORECASE if ignore_case else 0
    text = re.sub(
        pattern,
        lambda match: f"{color}{match.group()}{Fore.RESET}",
        match.line_text,
        flags=flags,
    ).strip()
    click.echo(f"  Line {match.line_number}: {text}")


def render_matches(
    pattern: str,
    match_dict: Mapping[Chapter, Mapping],
    color: AnsiFore,
    ignore_case: bool = False,
    describe: bool = False,
):
    for chapter, recipes in match_dict.items():
        click.echo(text_border(str(chapter)))
        for recipe, matches in recipes.items():
            render_recipe(recipe, describe)
            for match in matches:
                render_match(pattern, match, color, ignore_case)
            click.echo("")



import click
from .ls import ls
from .run import run
from .chapters import chapters
from .show import show
from .search import search


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
def main():
    """The CLI tool to find and display helpful Python recipes."""


main.add_command(ls)
main.add_command(run)
main.add_command(chapters)
main.add_command(show)
main.add_command(search)



"""
You would like to extend a closure with functions that allow the inner
variable to be accessed and modified.
"""
import sys


def sample():
    n = 0

    # Closure function
    def func():
        print("n =", n)

    # Accessor methods for n
    def get_n():
        return n

    def set_n(value):
        nonlocal n
        n = value

    # Attach as function attributes
    func.get_n = get_n
    func.set_n = set_n
    return func


class ClosureInstance:
    def __init__(self, locals=None):
        if locals is None:
            locals = sys._getframe(1).f_locals

        # Update the instance dict with callables
        self.__dict__.update(
            (key, value) for key, value in locals.items() if callable(value)
        )

    # Redirect special methods
    def __len__(self):
        return self.__dict__["__len__"]()


def Stack():
    items = []

    def push(item):
        items.append(item)

    def pop():
        return items.pop()

    def __len__():
        return len(items)

    return ClosureInstance()


def example_1():
    f = sample()
    f()
    f.set_n(10)
    f()
    print(f.get_n())


def example_2():
    s = Stack()
    print(s)
    s.push(10)
    s.push(20)
    s.push("Hello")
    n = len(s)
    print(n)
    for i in range(n):
        val = s.pop()
        print(val)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You need to format numbers for output, controlling the number of digits,
alignment, inclusion of thousands separator, and other details.
"""


def main():
    """
    General pattern for format is:
        '[<>^]?width[,]?(.digits)?'
    Where width and digits are integers and ? signifies optional parts.
    """
    x = 1234.56789

    print("Two decimal places of accuracy:")
    print(format(x, "0.2f"))

    print("\nRight justified in 10-chars, one-digit accuracy:")
    print(format(x, ">10.1f"))

    print("\nLeft justified")
    print(format(x, "<10.1f"))

    print("\nCentered")
    print(format(x, "^10.1f"))

    print("\nInclusion of thousand separator:")
    print(format(x, ","))
    print(format(x, "0,.1f"))

    print("\nExponential notation:")
    print(format(x, "e"))
    print(format(x, "0.2E"))


if __name__ == "__main__":
    main()



"""
You're writing code that navigates through a deeply nested tree structure
using the visitor pattern, but it blows up due to exceeding the recusion
limit. You'd like to eliminate the recusion, but keep the programming
style of the visitor pattern.
"""
import types
from .node import Node, NodeVisitor


def example_1():
    # Example:  Recursive implementation

    class UnaryOperator(Node):
        def __init__(self, operand):
            self.operand = operand

    class BinaryOperator(Node):
        def __init__(self, left, right):
            self.left = left
            self.right = right

    class Add(BinaryOperator):
        pass

    class Sub(BinaryOperator):
        pass

    class Mul(BinaryOperator):
        pass

    class Div(BinaryOperator):
        pass

    class Negate(UnaryOperator):
        pass

    class Number(Node):
        def __init__(self, value):
            self.value = value

    # A sample visitor class that evaluates expressions
    class Evaluator(NodeVisitor):
        def visit_Number(self, node):
            return node.value

        def visit_Add(self, node):
            return self.visit(node.left) + self.visit(node.right)

        def visit_Sub(self, node):
            return self.visit(node.left) - self.visit(node.right)

        def visit_Mul(self, node):
            return self.visit(node.left) * self.visit(node.right)

        def visit_Div(self, node):
            return self.visit(node.left) / self.visit(node.right)

        def visit_Negate(self, node):
            return -self.visit(node.operand)

    # 1 + 2*(3-4) / 5
    t1 = Sub(Number(3), Number(4))
    t2 = Mul(Number(2), t1)
    t3 = Div(t2, Number(5))
    t4 = Add(Number(1), t3)

    # Evaluate it
    e = Evaluator()
    print(e.visit(t4))  # Outputs 0.6

    # Blow it up

    a = Number(0)
    for n in range(1, 100000):
        a = Add(a, Number(n))

    try:
        print(e.visit(a))
    except RuntimeError as e:
        print(e)


def example_2():
    class UnaryOperator(Node):
        def __init__(self, operand):
            self.operand = operand

    class BinaryOperator(Node):
        def __init__(self, left, right):
            self.left = left
            self.right = right

    class Add(BinaryOperator):
        pass

    class Sub(BinaryOperator):
        pass

    class Mul(BinaryOperator):
        pass

    class Div(BinaryOperator):
        pass

    class Negate(UnaryOperator):
        pass

    class Number(Node):
        def __init__(self, value):
            self.value = value

    class Evaluator(NodeVisitor):
        def visit_Number(self, node):
            return node.value

        def visit_Add(self, node):
            yield (yield node.left) + (yield node.right)

        def visit_Sub(self, node):
            yield (yield node.left) - (yield node.right)

        def visit_Mul(self, node):
            yield (yield node.left) * (yield node.right)

        def visit_Div(self, node):
            yield (yield node.left) / (yield node.right)

        def visit_Negate(self, node):
            yield -(yield node.operand)

    # 1 + 2*(3-4) / 5
    t1 = Sub(Number(3), Number(4))
    t2 = Mul(Number(2), t1)
    t3 = Div(t2, Number(5))
    t4 = Add(Number(1), t3)

    # Evaluate it
    e = Evaluator()
    print(e.visit(t4))  # Outputs 0.6

    # Blow it up

    a = Number(0)
    for n in range(1, 100000):
        a = Add(a, Number(n))

    try:
        print(e.visit(a))
    except RuntimeError as e:
        print(e)


def example_3():
    class Node:
        pass

    class Visit:
        def __init__(self, node):
            self.node = node

    class NodeVisitor:
        def visit(self, node):
            stack = [Visit(node)]
            last_result = None
            while stack:
                try:
                    last = stack[-1]
                    if isinstance(last, types.GeneratorType):
                        stack.append(last.send(last_result))
                        last_result = None
                    elif isinstance(last, Visit):
                        stack.append(self._visit(stack.pop().node))
                    else:
                        last_result = stack.pop()
                except StopIteration:
                    stack.pop()
            return last_result

        def _visit(self, node):
            methname = "visit_" + type(node).__name__
            meth = getattr(self, methname, None)
            if meth is None:
                meth = self.generic_visit
            return meth(node)

        def generic_visit(self, node):
            raise RuntimeError("No {} method".format("visit_" + type(node).__name__))

    class UnaryOperator(Node):
        def __init__(self, operand):
            self.operand = operand

    class BinaryOperator(Node):
        def __init__(self, left, right):
            self.left = left
            self.right = right

    class Add(BinaryOperator):
        pass

    class Sub(BinaryOperator):
        pass

    class Mul(BinaryOperator):
        pass

    class Div(BinaryOperator):
        pass

    class Negate(UnaryOperator):
        pass

    class Number(Node):
        def __init__(self, value):
            self.value = value

    class Evaluator(NodeVisitor):
        def visit_Number(self, node):
            return node.value

        def visit_Add(self, node):
            yield (yield Visit(node.left)) + (yield Visit(node.right))

        def visit_Sub(self, node):
            yield (yield Visit(node.left)) - (yield Visit(node.right))

        def visit_Mul(self, node):
            yield (yield Visit(node.left)) * (yield Visit(node.right))

        def visit_Div(self, node):
            yield (yield Visit(node.left)) / (yield Visit(node.right))

        def visit_Negate(self, node):
            yield -(yield Visit(node.operand))

    # 1 + 2*(3-4) / 5
    t1 = Sub(Number(3), Number(4))
    t2 = Mul(Number(2), t1)
    t3 = Div(t2, Number(5))
    t4 = Add(Number(1), t3)

    # Evaluate it
    e = Evaluator()
    print(e.visit(t4))  # Outputs 0.6

    # Blow it up
    a = Number(0)
    for n in range(1, 100000):
        a = Add(a, Number(n))

    try:
        print(e.visit(a))
    except RuntimeError as e:
        print(e)


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You want to keep limited history of the last few items seen during iteration
or during some kind of processing.
"""

from collections import deque
from pathlib import Path


def search(lines, pattern, history=5):
    previous_lines = deque(maxlen=history)
    for line in lines:
        if pattern in line:
            yield line, previous_lines
        previous_lines.append(line)


def main():
    with open(Path(__file__).parent / "somefile.txt", "r") as f:
        for line, previous_lines in search(f, "python", 5):
            for prev_line in previous_lines:
                print(prev_line, end="")
            print(line, end="")
            print("-" * 20)


if __name__ == "__main__":
    main()



"""
You want an object to support customized formatting through the format()
function and string method.
"""

_formats = {
    "ymd": "{d.year}-{d.month}-{d.day}",
    "mdy": "{d.month}/{d.day}/{d.year}",
    "dmy": "{d.day}/{d.month}/{d.year}",
}


class Date:
    def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day

    def __format__(self, code):
        code = "ymd" if code == "" else code
        fmt = _formats[code]
        return fmt.format(d=self)


def main():
    d = Date(2012, 12, 21)
    print(d)
    print(format(d, "ymd"))
    print(format(d, "mdy"))
    print(format(d, "dmy"))
    print("The date is {date}".format(date=d))
    print("The date is {date:mdy}".format(date=d))
    print("The date is {date:dmy}".format(date=d))


if __name__ == "__main__":
    main()



"""
You want to implement new kinds of context managers for use with the 'with' statement.
"""

import time
from contextlib import contextmanager


def example_1():
    @contextmanager
    def timethis(label):
        start = time.time()
        try:
            yield
        finally:
            end = time.time()
            print("{}: {}".format(label, end - start))

    # Example use
    with timethis("counting"):
        n = 10000000
        while n > 0:
            n -= 1


def example_2():
    @contextmanager
    def list_transaction(orig_list):
        working = list(orig_list)
        yield working
        orig_list[:] = working

    # Example
    items = [1, 2, 3]
    with list_transaction(items) as working:
        working.append(4)
        working.append(5)
    print(items)
    try:
        with list_transaction(items) as working:
            working.append(6)
            working.append(7)
            raise RuntimeError("oops")
    except RuntimeError as e:
        print(e)

    print(items)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You need to read or write data in a file with gzip of bz2 compression.
"""
import gzip
import bz2
from pathlib import Path


THIS_DIR = Path(__file__).parent


def main():
    # Writing
    with gzip.open(THIS_DIR / "somefile.gz", "wt") as f:
        f.write("Hello World!")

    with bz2.open(THIS_DIR / "somfile.bz2", "wt") as f:
        f.write("Hello World!")

    # Reading
    with gzip.open(THIS_DIR / "somefile.gz", "rt") as f:
        text = f.read()
        print(text)

    with bz2.open(THIS_DIR / "somfile.bz2", "rt") as f:
        text = f.read()
        print(text)


if __name__ == "__main__":
    main()



"""
You need to unpack N elements from an iterable, but the iterable may
be longer than N elements, causing 'too many values to unpack' exception
"""


def drop_first_last(grades):
    assert len(grades) >= 3, "expect grades to be of length at least 3"
    _, *middle, _ = grades
    return sum(middle) / len(middle)


def main():
    grades = [30, 60, 74, 76, 78, 90]
    print("grades:", grades)
    print("avg:", drop_first_last(grades))

    record = ("Dave", "dave@example.com", "777-555-1212", "847-555-1212")
    print("record:", record)
    name, email, *phone_numbers = record
    print("name:", name)
    print("email:", email)
    print("phone numbers:", phone_numbers)


if __name__ == "__main__":
    main()



"""
You need to manipulate pathnames in order to find the base filename,
directory name, absolute path, and so on.
"""
from pathlib import Path


def main():
    path = Path("/Users/cadams/Data/data.csv")
    print(path)
    print(path.name)
    print(path.parent)
    print(Path("tmp").joinpath("data", path.name))

    path = Path("~/Data/data.csv")
    print(path)
    print(path.expanduser())
    print(path.suffix)


if __name__ == "__main__":
    main()



"""
You have a byte string and you need to unpack it into an integer value.
Alternatively, you need to convert a large integer back to a byte string.
"""


def example_1():
    data = b"\x00\x124V\x00x\x90\xab\x00\xcd\xef\x01\x00#\x004"
    print("data:", data)
    print("data as int (little):", int.from_bytes(data, "little"))
    print("data as int (big):   ", int.from_bytes(data, "big"))
    print("=" * 20)


def example_2():
    number = 94522842520747284487117727783387188
    print("number:", number)
    print("number as bytes (little):", number.to_bytes(16, "little"))
    print("number as bytes (big):", number.to_bytes(16, "big"))


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You want to create a dictionary, and you also want to control
the order of items when iterating or serializing.
"""
from collections import OrderedDict
import json


def main():
    d = OrderedDict()
    d["foo"] = 1
    d["bar"] = 2
    d["spam"] = 3
    d["grok"] = 4

    for key, val in d.items():
        print(key, val)

    print(json.dumps(d))


if __name__ == "__main__":
    main()



"""
You need to create an instance, but want to bypass the execution
of the __init__() method for some reeason.
"""
from time import localtime


class Date:
    def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day

    # Class method that bypasses __init__
    @classmethod
    def today(cls):
        d = cls.__new__(cls)
        t = localtime()
        d.year = t.tm_year
        d.month = t.tm_mon
        d.day = t.tm_mday
        return d


def main():
    d = Date.__new__(Date)
    print(d)
    print(hasattr(d, "year"))

    data = {"year": 2012, "month": 8, "day": 29}

    d.__dict__.update(data)
    print(d.year)
    print(d.month)

    d = Date.today()
    print(d.year, d.month, d.day)


if __name__ == "__main__":
    main()



import re
from pathlib import Path


def clean_text(text):
    num, text = text.split("_", maxsplit=1)
    return f"{int(num)}) {text.replace('_', ' ').capitalize()}"


def extract_leading_numbers(text):
    m = re.match(r"^\d+", text)
    if m:
        return int(m.group())


def text_border(text, symbol="=", side_symbol="=", padding=1):
    width = 2 * padding + len(text) + 2
    top = bottom = f"{symbol * width}"
    pad = " " * padding
    return f"{top}\n{side_symbol}{pad}{text}{pad}{side_symbol}\n{bottom}"


def tree(dir_path: Path, prefix: str = ""):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """
    # prefix components:
    space = "    "
    branch = "│   "
    # pointers:
    tee = "├── "
    last = "└── "

    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield prefix + pointer + path.name
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(path, prefix=prefix + extension)



"""
You've defined an anonymous function using lambda, but you also need
to capture the values of certain variables at the time of definition.
"""


def example_1():
    print("example 1")
    x = 10
    a = lambda y: x + y
    x = 20
    b = lambda y: x + y

    print(a(10))
    print(b(10))

    x = 15
    print(a(10))
    print()


def example_2():
    print("example 2")
    x = 10
    a = lambda y, x=x: x + y
    x = 20
    b = lambda y, x=x: x + y
    print(a(10))
    print(b(10))
    print()


def example_3():
    print("example 3")
    funcs = [lambda x, n=n: x + n for n in range(5)]
    for f in funcs:
        print(f(0))
    print()


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You want to read and write data over a serial port, typically to
interect wit some kind of harware device (e.g. a robot or sensor).
"""


def example_1():
    """For illustrative purposes only.
    Requires pyserial package.
        `pip install pyserial`
    """
    import serial

    ser = serial.Serial(
        "/dev/tty.usbmodem641",  # Device name varies
        baudrate=9600,
        bytesize=8,
        parity="N",
        stopbits=1,
    )

    ser.write(b"G1 X50 Y50 Y50\r\n")
    resp = ser.readline()
    print(resp)


def main():
    pass


if __name__ == "__main__":
    main()



"""
You need to perform calculations on large numerical datasets,
such as arrays or grids.
"""
import numpy as np


def main():
    # Lists
    x = [1, 2, 3, 4]
    y = [5, 6, 7, 8]
    print("x:", x, "y:", y)
    print("x * 2:", x * 2)
    print("x + y:", x + y)
    try:
        print("x + 10:", x + 10)
    except Exception as exc:
        print("x + 10:", exc)
    print()

    # NumPy arrays
    ax = np.array(x)
    ay = np.array(y)
    print("ax:", ax, "ay:", ay)
    print("ax * 2:", ax * 2)
    print("ax + ay:", ax + ay)
    print("ax + 10:", ax + 10)
    print("ax * ay:", ax * ay)
    print(np.sqrt(ax))
    print(np.cos(ax))
    print()

    # Grids
    m = np.zeros(shape=(1000, 1000), dtype=float)
    print(m)
    m += 10
    print(m)
    print(np.sin(m))

    # Slicing and changing subsections
    a = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
    print(a)
    print(a[1])
    print(a[:, 1])
    print(a[1:3, 1:3])
    a[1:3, 1:3] += 10
    print(a)

    # Broadcast row vector across an operation on all rows
    print(a + [100, 101, 102, 103])

    # Conditional assignment
    print(np.where(a < 10, a, 10))


if __name__ == "__main__":
    main()



"""
You want to redirect the output of the print() function to a file.
"""
from pathlib import Path


def main():
    with (Path(__file__).parent / "somefile.txt").open("wt") as f:
        print("hello world", file=f)


if __name__ == "__main__":
    main()



"""
You need to test whether or not a file or directory exists.
"""
from pathlib import Path


def main():
    for path in ["/tmp", __file__, "/totally/fake/path"]:
        print(path)
        print("exists:", Path(path).exists())
        print("is file:", Path(path).is_file())
        print("is dir:", Path(path).is_dir())
        print("is symlink:", Path(path).is_symlink())
        print()


if __name__ == "__main__":
    main()



import struct
import itertools
from pathlib import Path

polys = [
    [(1.0, 2.5), (3.5, 4.0), (2.5, 1.5)],
    [(7.0, 1.2), (5.1, 3.0), (0.5, 7.5), (0.8, 9.0)],
    [(3.4, 6.3), (1.2, 0.5), (4.6, 9.2)],
]


def write_polys(filename, polys):
    # Determine bounding box
    flattened = list(itertools.chain(*polys))
    min_x = min(x for x, y in flattened)
    max_x = max(x for x, y in flattened)
    min_y = min(y for x, y in flattened)
    max_y = max(y for x, y in flattened)

    with open(filename, "wb") as f:
        f.write(struct.pack("<iddddi", 0x1234, min_x, min_y, max_x, max_y, len(polys)))

        for poly in polys:
            size = len(poly) * struct.calcsize("<dd")
            f.write(struct.pack("<i", size + 4))
            for pt in poly:
                f.write(struct.pack("<dd", *pt))


# Call it with our polygon data
write_polys(str(Path(__file__).parent / "polys.bin"), polys)



"""
You want to read binary data directly into a mutable buffer without
any immediate copying. Perhaps you want to mutate the data in-place
and write it back out to a file.
"""
import os.path
from pathlib import Path


def read_into_buffer(filename):
    buf = bytearray(os.path.getsize(filename))
    with open(filename, "rb") as f:
        f.readinto(buf)
    return buf


def main():
    # Write sample file
    filename = Path(__file__).parent / "sample.bin"

    with filename.open("wb") as f:
        f.write(b"Hello World!")

    buf = read_into_buffer(filename)
    print(buf)
    print(buf[:5])
    buf[0:5] = b"hallo"
    print(buf)

    m1 = memoryview(buf)
    m2 = m1[-6:]
    print(m1, m2)
    m2[:] = b"WORLD!"
    print(buf)


if __name__ == "__main__":
    main()



"""
You want to encapsulate 'private' data on instances of a class,
but are concerned about Python's lack of access control.
"""


class A:
    def __init__(self):
        self._internal = 0  # An internal attribute
        self.public = 1  # A public attribute

    def public_method(self):
        """A public method"""
        print("public")

    def _internal_method(seld):
        """A private method"""
        print("private")


def main():
    a = A()
    print(a._internal)
    print(a.public)
    a._internal_method()
    a.public_method()


if __name__ == "__main__":
    main()



import types


class Node:
    pass


class NodeVisitor:
    def visit(self, node):
        stack = [node]
        last_result = None
        while stack:
            try:
                last = stack[-1]
                if isinstance(last, types.GeneratorType):
                    stack.append(last.send(last_result))
                    last_result = None
                elif isinstance(last, Node):
                    stack.append(self._visit(stack.pop()))
                else:
                    last_result = stack.pop()
            except StopIteration:
                stack.pop()
        return last_result

    def _visit(self, node):
        methname = "visit_" + type(node).__name__
        meth = getattr(self, methname, None)
        if meth is None:
            meth = self.generic_visit
        return meth(node)

    def generic_visit(self, node):
        raise RuntimeError("No {} method".format("visit_" + type(node).__name__))



"""
You want to replace HTML or XML entities such as &entity; or &#code;
with their corresponding text. Alternatively, you need to produce
text, but escape certain characters (e.g. <, >, &)
"""
import html
from xml.sax.saxutils import unescape


def main():
    s = 'Elements are writen as "<tag>text</tag>".'
    print(s)
    print(html.escape(s))
    print(html.escape(s, quote=False), end="\n\n")

    s = "Spicy Jalapeño"
    print(s)
    print(s.encode("ascii", errors="xmlcharrefreplace"), end="\n\n")

    s = "Spicy &quot;Jalape&#241;o&quot;"
    print(s)
    print(html.unescape(s), end="\n\n")

    t = "The prompt is &gt;&gt;&gt;"
    print(t)
    print(unescape(t))


if __name__ == "__main__":
    main()



"""
You are using exec() to execute a fragment of code in the scope of the
caller, but after execution, none of the results seem to be visible.
"""


def test():
    a = 13
    loc = locals()
    exec("b = a + 1")
    b = loc["b"]
    print(b)  # --> 14


def test1():
    x = 0
    exec("x += 1")
    print(x)  # --> 0


def test2():
    x = 0
    loc = locals()
    print("before:", loc)
    exec("x += 1")
    print("after:", loc)
    print("x =", x)


def test3():
    x = 0
    loc = locals()
    print(loc)
    exec("x += 1")
    print(loc)
    locals()
    print(loc)


def test4():
    a = 13
    loc = {"a": a}
    glb = {}
    exec("b = a + 1", glb, loc)
    b = loc["b"]
    print(b)


def main():
    print(":::: Running test()")
    test()

    print(":::: Running test1()")
    test1()

    print(":::: Running test2()")
    test2()

    print(":::: Running test3()")
    test3()

    print(":::: Running test4()")
    test4()


if __name__ == "__main__":
    main()



"""
You need to check the start or end of a string for specific text pattern
such as file extensions, url schemes etc.
"""

from urllib.request import urlopen

filename = "spam.txt"
url = "http://www.python.org"


def example_1():
    print(f'Checking if filename {filename} ends with the string ".txt"')
    print(filename.endswith(".txt"), end="\n\n")
    print(f'Checking if filename {filename} starts with "file:"')
    print(filename.startswith("file:"), end="\n\n")


def example_2():
    print(f"Checking if {url} starts wih either (http:|https|s3:|ftp:)")
    print(url.startswith(("http:", "https:", "s3:", "ftp:")))


def read_data(name):  # pragma: no cover
    if name.startswith(("http:", "https:", "ftp:")):
        return urlopen(name).read()
    else:
        with open(name, "r") as f:
            return f.read()


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You're trying to match a text pattern using regular expressions, but
it is identifying the longest possible match of a pattern. Instead,
you would like to change it to find the shortest possible match.
"""

import re


def main():
    str_pat = re.compile(r"\"(.*)\"")
    text1 = 'computer says "no."'
    matches = str_pat.findall(text1)
    print("pattern:", str_pat)
    print("text:", text1)
    print("matches:", matches)

    text2 = 'computer says "no." Phone says "yes."'
    matches = str_pat.findall(text2)
    print("text2:", text2)
    print("matches:", matches)
    print()

    str_pat = re.compile(r"\"(.*?)\"")
    matches = str_pat.findall(text2)
    print("pattern:", str_pat)
    print("text2:", text2)
    print("matches:", matches)


if __name__ == "__main__":
    main()



"""
You need to select, insert or delete rows in a relational database.
"""
import sqlite3


def main():
    stocks = [
        ("GOOG", 100, 490.1),
        ("AAPL", 50, 545.75),
        ("FB", 150, 7.45),
        ("HPQ", 75, 33.2),
    ]
    print("stocks:", stocks)

    with sqlite3.connect(":memory:") as db:
        print(db)

        c = db.cursor()
        c.execute(
            """
            CREATE TABLE portfolio
            (symbol text,
            shares integer,
            price real)
            """
        )
        db.commit()

        c.executemany(
            """
            INSERT INTO portfolio VALUES (?, ?, ?)
            """,
            stocks,
        )

        for row in db.execute("SELECT * FROM portfolio"):
            print(row)


if __name__ == "__main__":
    main()



"""
You want to invoke a method in a parent class in place of a method
that has been overridden in a subclass.
"""


class A:
    def spam(self):
        print("A.spam")


class B(A):
    def spam(self):
        print("B.spam")
        super().spam()


class C:
    def __init__(self):
        self.x = 0


class D(C):
    def __init__(self):
        super().__init__()
        self.y = 1


def example_1():
    a = A()
    b = B()
    a.spam()
    b.spam()


def example_2():
    c = C()
    d = D()
    print(c.x)
    print(d.x)
    print(d.y)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



import sys
import click
from colorama import Fore, init
from pyrecipes.cookbook import cookbook
from pyrecipes.utils.cli import get_n_matches, render_matches

init(autoreset=True)

COLORS = {
    "black": Fore.BLACK,
    "red": Fore.RED,
    "green": Fore.GREEN,
    "yellow": Fore.YELLOW,
    "blue": Fore.BLUE,
    "magenta": Fore.MAGENTA,
    "cyan": Fore.CYAN,
    "white": Fore.WHITE,
    "none": Fore.RESET,
}


@click.command()
@click.argument("pattern", type=str)
@click.option(
    "--color",
    type=click.Choice(COLORS.keys(), case_sensitive=False),
    default="red",
    help="Change the color used to highlight matches. The default is red.",
)
@click.option(
    "-i",
    "--ignore-case",
    is_flag=True,
    default=False,
    help="Make the search case-insensitive",
)
@click.option(
    "-c", "--count-only", is_flag=True, help="Return the count of matches only."
)
@click.option(
    "-d",
    "--describe",
    is_flag=True,
    help="Shows descriptions of the recipes",
)
def search(pattern, color, ignore_case, count_only, describe):
    """Search the recipes for a pattern

    \b
    - RegEx patterns are supported.
    - Searches are case-sensitive by default. Use the `-i` flag to make searches case-insensitive.
    - Use the `-c` flag to display count of matches.

    """
    color = COLORS.get(color)
    match_dict = cookbook.search(pattern, ignore_case)

    click.echo(f"Found {get_n_matches(match_dict):,} matches")

    if count_only:
        sys.exit(0)

    if match_dict:
        render_matches(pattern, match_dict, color, ignore_case, describe)



"""
You need to search for an possible replace text in a case-insensitive manner.
"""
import re


def matchcase(word: str):
    def replace(m):
        text = m.group()
        if text.isupper():
            return word.upper()
        elif text.islower():
            return word.lower()
        elif text[0].isupper():
            return word.capitalize()
        else:
            return word

    return replace


def main():
    text = "UPPER PYTHON, lower python, Mixed Python"
    print("text:", text)

    matches = re.findall("python", text, flags=re.IGNORECASE)
    print("matches:", matches)

    print("replace:", re.sub("python", "snake", text, flags=re.IGNORECASE))
    print(
        "replace (maintaining case):",
        re.sub("python", matchcase("snake"), text, flags=re.IGNORECASE),
    )


if __name__ == "__main__":
    main()



"""
You want to feed a text or binary string to code that's been written
to operate on file-like objects instead.
"""

import io


def main():
    s = io.StringIO()
    s.write("Hello world\n")
    print("this is a test", file=s)
    print(s.getvalue())

    s.seek(0)
    print(s.read(4))
    print(s.read())


if __name__ == "__main__":
    main()



"""
You want to apply a decorator to a class or static method.
"""

import time
from functools import wraps


# A simple decorator
def timethis(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        r = func(*args, **kwargs)
        end = time.time()
        print(end - start)
        return r

    return wrapper


# Class illustrating application of the decorator to different kinds of methods
class Spam:
    @timethis
    def instance_method(self, n):
        print(self, n)
        while n > 0:
            n -= 1

    @classmethod
    @timethis
    def class_method(cls, n):
        print(cls, n)
        while n > 0:
            n -= 1

    @staticmethod
    @timethis
    def static_method(n):
        print(n)
        while n > 0:
            n -= 1


def main():
    s = Spam()
    s.instance_method(10000000)
    Spam.class_method(10000000)
    Spam.static_method(10000000)


if __name__ == "__main__":
    main()



"""
You need to parse an XML document, but it's using XML namespaces.
"""
from pathlib import Path
from xml.etree.ElementTree import parse


class XMLNamespaces:
    def __init__(self, **kwargs):
        self.namespaces = {}
        for name, uri in kwargs.items():
            self.register(name, uri)

    def register(self, name, uri):
        self.namespaces[name] = "{" + uri + "}"

    def __call__(self, path):
        return path.format_map(self.namespaces)


def main():
    with open(Path(__file__).parent / "sample.xml", "r") as f:
        doc = parse(f)
    ns = XMLNamespaces(html="http://www.w3.org/1999/xhtml")

    e = doc.find(ns("content/{html}html"))
    print(e)

    text = doc.findtext(ns("content/{html}html/{html}head/{html}title"))
    print(text)


if __name__ == "__main__":
    main()



"""
You want an instance to delegate attrbute access to an internally
held instance, possibly as an alternative to inheritance or in
order to implement a proxy.
"""


class A:
    def spam(self, x):
        print(f"{self}.spam:", x)

    def foo(self):
        print(f"{self}.foo called")

    def __repr__(self):
        return f"{self.__class__.__name__}"


class B:
    # If only a small number of methods to delegate
    def __init__(self):
        self._a = A()

    def spam(self, x):
        # Delegate to the internal self._a instance
        self._a.spam(x)

    def foo(self):
        # Delegate to the internal self._a instance
        self._a.foo()

    def bar(self):
        print(f"{self}.bar called")

    def __repr__(self):
        return f"{self.__class__.__name__}"


class C:
    # Alternative approach using __getattr__
    def __init__(self):
        self._a = A()

    def __getattr__(self, name):
        return getattr(self._a, name)

    def __repr__(self):
        return f"{self.__class__.__name__}"


def main():
    b = B()
    print(b)
    b.bar()
    b.foo()
    b.spam(1)

    c = C()
    print(c)
    c.foo()
    c.spam(1)


if __name__ == "__main__":
    main()



"""
You have code organised as a package and want to import a submodule
from one of the other package submodules without hardcoding the
package name into the import statement.
"""

import textwrap
from pathlib import Path
from pyrecipes.utils.text import tree


def main():
    print(
        textwrap.dedent(
            """
            Suppose you have a package with the following structure:

            """
        )
    )

    for line in tree(Path(__file__).parent / "src"):
        print(line)

    print(
        textwrap.dedent(
            """
            You can use relative import statements to control the imports between these modules.

            Suppose for example within the module 'my_package.A.spam' you need to import both
            'my_package.A.grok' and 'my_package.B.bar', you can use the following statements:

            # my_package/A/spam.py
            from . import grok
            from ..B import bar
            """
        )
    )


if __name__ == "__main__":
    main()



"""
When creating instances of a class, you want to return a cached reference
to a previous instance created with the same arguments (if any).
"""


def example_1():
    # Simple example
    class Spam:
        def __init__(self, name):
            self.name = name

    # Caching support
    import weakref

    _spam_cache = weakref.WeakValueDictionary()

    def get_spam(name):
        if name not in _spam_cache:
            s = Spam(name)
            _spam_cache[name] = s
        else:
            s = _spam_cache[name]
        return s

    a = get_spam("foo")
    b = get_spam("bar")
    print("a is b:", a is b)
    c = get_spam("foo")
    print("a is c:", a is c)
    print()


def example_2():
    import weakref

    class CachedSpamManager:
        def __init__(self):
            self._cache = weakref.WeakValueDictionary()

        def get_spam(self, name):
            if name not in self._cache:
                s = Spam(name)
                self._cache[name] = s
            else:
                s = self._cache[name]
            return s

    class Spam:
        def __init__(self, name):
            self.name = name

    Spam.manager = CachedSpamManager()

    def get_spam(name):
        return Spam.manager.get_spam(name)

    a = get_spam("foo")
    b = get_spam("bar")
    print("a is b:", a is b)
    c = get_spam("foo")
    print("a is c:", a is c)
    print()


def example_3():
    # Example involving new and some of its problems
    import weakref

    class Spam:
        _spam_cache = weakref.WeakValueDictionary()

        def __new__(cls, name):
            if name in cls._spam_cache:
                return cls._spam_cache[name]
            else:
                self = super().__new__(cls)
                cls._spam_cache[name] = self
                return self

        def __init__(self, name):
            print("Initializing Spam")
            self.name = name

    print("This should print 'Initializing Spam' twice")
    s = Spam("Dave")
    t = Spam("Dave")
    print(s is t)


def main():
    example_1()
    example_2()
    example_3()


if __name__ == "__main__":
    main()



"""
You've written a function, but would like to attach some additional
information to the arguments so that others know more about how a
function is supposed to be used.
"""

import pydoc


def add(x: int, y: int):
    """Adds 2 integers"""
    return x + y


def main():
    x = 2
    y = 3
    print(f"x: {x}, y: {y}")
    print(pydoc.render_doc(add))
    print(add(x, y))
    print(add.__annotations__)


if __name__ == "__main__":
    main()



"""
You need to extract data from a huge XML document using as little memory as possible.
"""
from xml.etree.ElementTree import iterparse
from pathlib import Path
from collections import Counter


def parse_and_remove(filename, path):
    path_parts = path.split("/")
    doc = iterparse(filename, ("start", "end"))
    # Skip the root element
    next(doc)

    tag_stack = []
    elem_stack = []
    for event, elem in doc:
        if event == "start":
            tag_stack.append(elem.tag)
            elem_stack.append(elem)
        elif event == "end":
            if tag_stack == path_parts:
                yield elem
                elem_stack[-2].remove(elem)
            try:
                tag_stack.pop()
                elem_stack.pop()
            except IndexError:
                pass


def main():
    # Find zip code with most potholes
    potholes_by_zip = Counter()
    file = Path(__file__).parent / "potholes.xml"

    data = parse_and_remove(file, "row/row")
    for pothole in data:
        potholes_by_zip[pothole.findtext("zip")] += 1

    for zipcode, num in potholes_by_zip.most_common():
        print(zipcode, num)


if __name__ == "__main__":
    main()



"""
You'd like to be able to compare instances of you class using
the standard comparison operators (e.g. >=, !=, == etc.), but
without having to write a lot of special methods.
"""
from functools import total_ordering


class Room:
    def __init__(self, name, length, width):
        self.name = name
        self.length = length
        self.width = width
        self.square_feet = self.length * self.width


@total_ordering
class House:
    def __init__(self, name, style):
        self.name = name
        self.style = style
        self.rooms = list()

    @property
    def living_space_footage(self):
        return sum(r.square_feet for r in self.rooms)

    def add_room(self, room):
        self.rooms.append(room)

    def __str__(self):
        return "{}: {} square foot {}".format(
            self.name, self.living_space_footage, self.style
        )

    def __eq__(self, other):
        return self.living_space_footage == other.living_space_footage

    def __lt__(self, other):
        return self.living_space_footage < other.living_space_footage


def main():
    # Build a few houses, and add rooms to them.
    h1 = House("h1", "Cape")
    h1.add_room(Room("Master Bedroom", 14, 21))
    h1.add_room(Room("Living Room", 18, 20))
    h1.add_room(Room("Kitchen", 12, 16))
    h1.add_room(Room("Office", 12, 12))

    h2 = House("h2", "Ranch")
    h2.add_room(Room("Master Bedroom", 14, 21))
    h2.add_room(Room("Living Room", 18, 20))
    h2.add_room(Room("Kitchen", 12, 16))

    h3 = House("h3", "Split")
    h3.add_room(Room("Master Bedroom", 14, 21))
    h3.add_room(Room("Living Room", 18, 20))
    h3.add_room(Room("Office", 12, 16))
    h3.add_room(Room("Kitchen", 15, 17))
    houses = [h1, h2, h3]

    print("Is h1 bigger than h2?", h1 > h2)  # prints True
    print("Is h2 smaller than h3?", h2 < h3)  # prints True
    print("Is h2 greater than or equal to h1?", h2 >= h1)  # prints False
    print("Which one is biggest?", max(houses))  # prints 'h3: 1101 square foot Split'
    print("Which is smallest?", min(houses))  # prints 'h2: 846 square foot Ranch'


if __name__ == "__main__":
    main()



"""
You want to pick random items out of a sequence or generate random numbers.
"""
import random


def main():
    random.seed(0)
    values = [1, 2, 3, 4, 5, 6]
    print("values:", values)

    print("\npick a random value from values")
    for _ in range(5):
        print(random.choice(values))

    print("\npick a random sample of values from values")
    for _ in range(5):
        print(random.sample(values, 3))

    print("\nrandomly shuffle values")
    for _ in range(5):
        random.shuffle(values)
        print(values)

    print("\npick a random integer")
    for _ in range(5):
        print(random.randint(0, 10))

    print("\npick a random floating-point between 0-1")
    for _ in range(5):
        print(random.random())

    print("\nget N random bits expressed as an integer")
    for _ in range(5):
        print(random.getrandbits(200))


if __name__ == "__main__":
    main()



"""
You need to perform matrix and linear algebra operations, such as
matrix multiplication, finding determinants, solving linear equations
and so on.
"""
import numpy as np


def main():
    m = np.array([[1, -2, 3], [0, 4, 5], [7, 8, -9]])
    print("m:", m, sep="\n", end="\n\n")
    print("m transposed:", m.T, sep="\n", end="\n\n")
    print("m inverse:", np.linalg.inv(m), sep="\n", end="\n\n")

    v = np.array([[2], [3], [4]])
    print("v:", v)
    print("m * v:", m * v, end="\n\n")

    print("determinant:", np.linalg.det(m))
    print("eigenvalues:", np.linalg.eigvals(m))

    x = np.linalg.solve(m, v)
    print("solve for x in mx = v:\n", x)
    print("m * x:\n", m * x)


if __name__ == "__main__":
    main()



"""
You have the name of a method you that want to call on an object stored
as a string and you want to execute the method.
"""
import math


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return "Point({!r:},{!r:})".format(self.x, self.y)

    def distance(self, x, y):
        return math.hypot(self.x - x, self.y - y)


def main():
    p = Point(2, 3)

    # Method 1 : Use getattr
    d = getattr(p, "distance")(0, 0)  # Calls p.distance(0, 0)
    print(d)

    # Method 2: Use methodcaller
    import operator

    d = operator.methodcaller("distance", 0, 0)(p)
    print(d)

    # Application in sorting
    points = [
        Point(1, 2),
        Point(3, 0),
        Point(10, -3),
        Point(-5, -7),
        Point(-1, 8),
        Point(3, 2),
    ]

    # Sort by distance from origin (0, 0)
    points.sort(key=operator.methodcaller("distance", 0, 0))
    for p in points:
        print(p)


if __name__ == "__main__":
    main()



"""
You launched a thread, but want to know when it actually starts running.
"""

import time
import threading
from threading import Thread, Event


def countdown(n, started_event):
    print("countdown started")
    started_event.set()
    while n > 0:
        print(f"T-minus {n}")
        n -= 1
        time.sleep(1)


def example_1():
    # Create the event object used to signal startup
    started_event = Event()

    # Launch the thread and pass the startup event
    print("Laumching countdown")
    t = Thread(target=countdown, args=(10, started_event), daemon=True)
    t.start()

    # Wait for the thread to start
    started_event.wait()
    print("Countdown is running")


def example_2():  # pragma: no cover
    class PeriodicTimer:
        def __init__(self, interval):
            self._interval = interval
            self._flag = 0
            self._cv = threading.Condition()

        def start(self):
            t = threading.Thread(target=self.run, daemon=True)
            t.start()

        def run(self):
            """
            Run the timer and notify waiting threads after each interval
            """
            while True:
                time.sleep(self._interval)
                with self._cv:
                    self._flag ^= 1
                    self._cv.notify_all()

        def wait_for_tick(self):
            """
            Wait for the next tick of the timer
            """
            with self._cv:
                last_flag = self._flag
                while last_flag == self._flag:
                    self._cv.wait()

    # Example use of the timer
    ptimer = PeriodicTimer(5)
    ptimer.start()

    # Two threads that synchronize on the timer
    def countdown(nticks):
        while nticks > 0:
            ptimer.wait_for_tick()
            print("T-minus", nticks)
            nticks -= 1

    def countup(last):
        n = 0
        while n < last:
            ptimer.wait_for_tick()
            print("Counting", n)
            n += 1

    threading.Thread(target=countdown, args=(10,)).start()
    threading.Thread(target=countup, args=(5,)).start()


def main():
    example_1()
    # example_2()


if __name__ == "__main__":
    main()



import pytest

from pyrecipes.chapter import Chapter
from pyrecipes.errors import RecipeNotFoundError
from pyrecipes.recipe import Recipe


@pytest.fixture
def chapter(recipe_root_dir):
    yield Chapter(recipe_root_dir / "01_test_chapter")


def test_Chapter_init(chapter, recipe_root_dir):
    assert chapter.chapter_dir == recipe_root_dir / "01_test_chapter"
    assert chapter.number == 1
    assert chapter.name == "01_test_chapter"
    assert chapter.recipes.keys() == {1, 2, 3}
    assert str(chapter) == "1) Test Chapter"
    assert chapter.__repr__() == "Chapter(number=1, name=01_test_chapter, recipes=3)"
    assert chapter.size == 3


def test_Chapter_indexing_found_returns_recipe(chapter):
    recipe = chapter[1]
    assert isinstance(recipe, Recipe)
    assert recipe.number == 1


def test_Chapter_indexing_not_found_raises_RecipeNotFoundError(chapter):
    with pytest.raises(RecipeNotFoundError, match="Recipe 1.200 not found"):
        _ = chapter[200]


def test_Chapter_iterating(chapter):
    expected_recipe_numbers = [1, 2, 3]
    expected_recipe_names = ["01_test_recipe", "02_test_recipe", "03_test_recipe"]

    for recipe, expected_number, expected_name in zip(
        chapter, expected_recipe_numbers, expected_recipe_names
    ):
        assert recipe.number == expected_number
        assert recipe.name == expected_name



"""
You want to write a decorator that adds an extra argument to the calling
signature of the wrapped function. However, the added argument can't
interfere with the existing calling conventions of the function.
"""

from functools import wraps


def optional_debug(func):
    @wraps(func)
    def wrapper(*args, debug=False, **kwargs):
        if debug:
            print("Calling", func.__name__)
        return func(*args, **kwargs)

    return wrapper


@optional_debug
def spam(a, b, c):
    print(a, b, c)


def main():
    spam(1, 2, 3)
    spam(1, 2, 3, debug=True)


if __name__ == "__main__":
    main()



"""
You're trying to match a block of text using a regular expression,
but you need the match to span multiple lines.
"""

import re


def main():
    comment = re.compile(r"/\*(.*?)\*/")
    text1 = "/* This is a comment */"
    text2 = """/* This is a\nmultiline comment */"""
    match1 = comment.findall(text1)
    match2 = comment.findall(text2)

    print("pattern:", comment)
    print("text1:", text1)
    print("text2:", text2)
    print("match1:", match1)
    print("match2:", match2)
    print()

    comment = re.compile(r"/\*((?:.|\n)*?)\*/")
    text1 = "/* This is a comment */"
    text2 = """/* This is a\nmultiline comment */"""
    match1 = comment.findall(text1)
    match2 = comment.findall(text2)

    print("pattern:", comment)
    print("text1:", text1)
    print("text2:", text2)
    print("match1:", match1)
    print("match2:", match2)


if __name__ == "__main__":
    main()



import sys
import pytest
import shlex
from pyrecipes.cli import main


def patch_argv(monkeypatch, command):
    args = shlex.split(command)
    monkeypatch.setattr(sys, "argv", args)
    return monkeypatch


def call_main(capsys):
    with pytest.raises(SystemExit) as exc:
        main()
    out, err = capsys.readouterr()
    return exc, out, err


@pytest.mark.parametrize("command", ["recipes", "recipes -h", "recipes --help"])
def test_no_command_prints_help(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, out, err = call_main(capsys)

    assert exc.value.code == 0
    assert "The CLI tool to find and display helpful Python recipes" in out
    assert "Usage: " in out
    assert err == ""


@pytest.mark.parametrize("command", ["recipes chapters", "recipes chapters -h"])
def test_chapter(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, _, err = call_main(capsys)

    assert exc.value.code == 0
    assert err == ""


@pytest.mark.parametrize(
    "command",
    ["recipes ls", "recipes ls 1", "recipes ls -d", "recipes ls -h", "recipes ls 1000"],
)
def test_ls(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, _, err = call_main(capsys)

    assert exc.value.code == 0
    assert err == ""


@pytest.mark.parametrize(
    "command", ["recipes show -h", "recipes show 1 3", "recipes show 1 1000"]
)
def test_show(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, _, err = call_main(capsys)

    assert exc.value.code == 0
    assert err == ""


@pytest.mark.parametrize(
    "command", ["recipes run 1 3", "recipes run -h", "recipes run 1 1000"]
)
def test_run(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, _, err = call_main(capsys)

    assert exc.value.code == 0
    assert err == ""


@pytest.mark.parametrize(
    "command",
    [
        "recipes search test",
        "recipes search -h",
        "recipes search foobar",
        "recipes search test -c",
        "recipes search -i test",
    ],
)
def test_search(monkeypatch, capsys, command):
    monkeypatch = patch_argv(monkeypatch, command)
    exc, _, err = call_main(capsys)

    assert exc.value.code == 0
    assert err == ""



"""
You have a callable that you would like to use with some other Python
code, possibly as a callback function of handler, but it takes too
many arguments and causes an exception when called.
"""
import math
from functools import partial


def spam(a, b, c, d):
    print(a, b, c, d)


def distance(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    return math.hypot(x2 - x1, y2 - y1)


def example_1():
    s1 = partial(spam, 1)
    s1(2, 3, 4)
    s1(4, 5, 6)

    s2 = partial(spam, d=42)
    s2(1, 2, 3)
    s2(4, 5, 5)

    s3 = partial(spam, 1, 2, d=42)
    s3(3)
    s3(4)
    s3(5)


def example_2():
    points = [
        (1, 2),
        (3, 4),
        (5, 6),
        (7, 8),
    ]
    print("points", points)
    pt = (4, 3)
    print("pt:", pt)
    points.sort(key=partial(distance, pt))
    print("sorted:", points)


def main():
    example_1()
    example_2()


if __name__ == "__main__":
    main()



"""
You have long strings that you want to reformat so that they fill a
user-specified number of columns.
"""
import textwrap
import os


def main():
    s = (
        "Look into my eyes, look into my eyes, the eyes, the eyes, "
        "the eyes, not around the eyes, don't look around the eyes, "
        "you're under"
    )
    print(s)

    print(textwrap.fill(s, 70))
    print(textwrap.fill(s, 40))
    print(textwrap.fill(s, 40, initial_indent="    "))
    print(textwrap.fill(s, 40, subsequent_indent="    "))

    try:
        columns = os.get_terminal_size().columns
        print("terminal size:", columns)
    except OSError:
        print("no terminal")


if __name__ == "__main__":
    main()
